︠27cf9449-f343-46ba-9675-619664e2b8d3i︠
%md
## Press open to interact with this worksheet.
Once in interactive mode, use the mouse to click or highlight a block. Then use the "run" button on top or "shift + enter" to execute the block.

To make a new block click on one of the horizontal bars. Here you can add Sage code or type in comments.

︡57874d4e-429e-43e4-8c02-ac3782ef351f︡{"done":true,"md":"## Press open to interact with this worksheet.\nOnce in interactive mode, use the mouse to click or highlight a block. Then use the \"run\" button on top or \"shift + enter\" to execute the block.\n\nTo make a new block click on one of the horizontal bars. Here you can add Sage code or type in comments."}
︠46837221-de83-4b76-88b3-2ff3f98a16cfi︠
%md
In the following worksheet we demonstrate how to compute the class group of a quadratic field as well as decompose ideals over a set of generators for the class group.
︡1a50d19b-d257-403a-a434-252afbd09c9f︡{"done":true,"md":"In the following worksheet we demonstrate how to compute the class group of a quadratic field as well as decompose ideals over a set of generators for the class group."}
︠eebffd8b-fbdc-4714-a9da-23d39ab446dcs︠

# In the following, consider the quadratic field with fundamental discriminant -1363.
D = -1363
K = QuadraticField(D)
K

︡a3d3393d-e6e8-460e-af73-e885913a90df︡{"stdout":"Number Field in a with defining polynomial x^2 + 1363 with a = 36.91882988394947?*I\n"}︡{"done":true}
︠5769b884-30a0-4433-b45e-19021e652b0ds︠

# Ideal classes of O_K can be represented with reduced quadratic forms of discriminant -1363.

f = BinaryQF([7, 3, 49])
f.is_reduced()
f.discriminant()    # verify the form has the right discriminant

︡a547cd6f-c181-4845-8975-04a9458cecaf︡{"stdout":"True\n"}︡{"stdout":"-1363\n"}︡{"done":true}
︠df0d93b3-fe44-41c1-ba17-84a9078bcb81s︠

# Composition of forms is denoted by multiplication, but the result is not necessarily reduced.

f*f
(f*f).reduced_form()

︡ee572700-aa8d-4c90-baf3-ed09909719f9︡{"stdout":"49*x^2 + 3*x*y + 7*y^2\n"}︡{"stdout":"7*x^2 - 3*x*y + 49*y^2\n"}︡{"done":true}
︠1a559b7a-5828-4b76-969a-3e2ea950906es︠

# If f = (a, b, c) is a reduced form, then its inverse is (a, -b, c)

g = BinaryQF([7, -3, 49])
g.is_reduced()

I = f*g
I.reduced_form()

︡37c97df9-c33e-4b8d-b1dc-ae0be51849d8︡{"stdout":"True\n"}︡{"stdout":"x^2 + x*y + 341*y^2\n"}︡{"done":true}
︠56445621-c714-4ca7-94d4-c318657fd253s︠

# The split prime ideals of K with norm less than Minkowski's bound generate the class group of K.
B = floor(K.minkowski_bound())
B

︡2db39603-9fff-48f5-93b7-5c67bc6be511︡{"stdout":"23\n"}︡{"done":true}
︠8d3546ab-c16b-4d25-9f3d-dda676f389edi︠
%md
### Q1. What are the ramified primes of K? Why do we not need to consider them when computing the class group?
︡b43fb381-7b01-4d45-9be3-8933edf44185︡{"done":true,"md":"### Q1. What are the ramified primes of K? Why do we not need to consider them when computing the class group?"}
︠93901f85-0060-44c4-a74f-ddd3e0c91a32i︠
%md
### Q2. What is the ideal class of an inert prime? Why do we not need to consider them when computing the class group?
︡dc1ec714-fee7-4d0b-9cdb-e9c0fdde607f︡{"done":true,"md":"### Q2. What is the ideal class of an inert prime? Why do we not need to consider them when computing the class group?"}
︠e096c4c9-07f4-4baf-8702-37adc94810abs︠

# We can construct the forms corresponding to reduced prime ideals. This is the prime form above 7, with its conjugate given by its inverse.

p = 7
R = Integers(2*p)

a = p
b = int(sqrt(R(D)))        # square root of D mod 7
c = (b^2 - D)/(4*a)

f = BinaryQF([a, b, c])
f = f.reduced_form()
f

f.discriminant()

︡d1ac8a8d-aed6-4684-b74d-4266f5316660︡{"stdout":"7*x^2 + 3*x*y + 49*y^2\n"}︡{"stdout":"-1363\n"}︡{"done":true}
︠46c1e2b1-2c66-4b0e-8dd6-d7b5e6d9b59ei︠

%md
### Q4. Find the binary quadratic forms f1, f2, ..., fn corresponding to the split prime ideals with norm less than B.

︡f621648f-1c31-46dd-8b99-d5fc2ab842a6︡{"done":true,"md":"### Q4. Find the binary quadratic forms f1, f2, ..., fn corresponding to the split prime ideals with norm less than B."}
︠3e3d211f-6c88-4cab-ab8f-4fa655b5ad27s︠

# We now need relations of the form f1^x1 * f2^x2 * ... * fn^xn = I = (1, 1, 341).
# Since exponentiation of forms is not natively supported here is a (not efficient) function to help:

id = BinaryQF([1, 1, 341])

def inv(f):
    g = f.reduced_form()
    return BinaryQF([g[0], -g[1], g[2]]).reduced_form()

def ex(f, n):
    g = id
    for i in range(abs(n)):
        g *= f
    g = g.reduced_form()
    if n < 0:
        g = inv(g)
    return g

inv(f)
ex(f, 2) == (f * f).reduced_form()


︡524a0588-f7d4-4113-ac1f-5adc4a974cdf︡{"stdout":"7*x^2 - 3*x*y + 49*y^2\n"}︡{"stdout":"True\n"}︡{"done":true}
︠6101b0e6-af7a-4e8f-ac72-f16e7a033b25i︠
%md
### Q5. Write a function that generates n random numbers in [-10, 10]. Look for relations of the form f1^x1 * f2^x2 * ... * fn^xn = I = (1, 1, 341).
︡73b9d865-b984-46e9-961d-47866d443ac8︡{"done":true,"md":"### Q5. Write a function that generates n random numbers in [-10, 10]. Look for relations of the form f1^x1 * f2^x2 * ... * fn^xn = I = (1, 1, 341)."}
︠586c20d2-e746-4877-9e7a-19fb02df89b2s︠


# Here is a simple example of working with matrices in Sage, especially adding rows to a matrix:

M = matrix([[1, 2, 4], [5, 4, 1], [2, 4, 3]])

rows = M.rows()
rows += [(5, 6, 7)]

matrix(rows)

︡bc224c3a-1d59-4af3-aa9a-3b05d5127fea︡{"stdout":"[1 2 4]\n[5 4 1]\n[2 4 3]\n[5 6 7]\n"}︡{"done":true}
︠71bfee18-0237-44f6-a786-b7e1642b24e0i︠

%md
### Q6. Construct the matrix with row vectors given by (x1, x2, ..., xn). Continue searching for relations as in Q3 until the matrix has full rank.
︡4e0a9195-6883-49fe-99c8-99e0337111d1︡{"done":true,"md":"### Q6. Construct the matrix with row vectors given by (x1, x2, ..., xn). Continue searching for relations as in Q3 until the matrix has full rank."}
︠d58dccbd-6b8d-420d-9ca0-f9c8eba0bcads︠

# Recall that the Hermite normal form of this matrix gives the structure of the class group of K.
# We can create the (upper triangular!) HNF by:

M.hermite_form()
print()

# and the Smith normal form by:

M.smith_form()

︡0fa4c2e5-6071-4886-83da-7d2de89ac73f︡{"stdout":"[1 2 4]\n[0 6 4]\n[0 0 5]\n"}︡{"stdout":"\n"}︡{"stdout":"([ 1  0  0]\n[ 0  1  0]\n[ 0  0 30], [  0   0   1]\n[  0   1   0]\n[  1 -25 -13], [  3   7   8]\n[ -5 -11 -13]\n[  5  10  12])\n"}︡{"done":true}
︠a6f48d7b-2139-44c4-899d-0cc5c3e353f0i︠

%md
### Q7. What is the structure of the class group? (Hint: the class number is less than 12. If you still aren't there, try adding more relations.)
︡29034339-bed4-4d1a-9a8e-4859fa51d25b︡{"done":true,"md":"### Q7. What is the structure of the class group? (Hint: the class number is less than 12. If you still aren't there, try adding more relations.)"}
︠ad0c8633-c0c7-4176-9ab7-f4c0ef74bcb3i︠
%md
### Q8. Find elements of each possible order in the class group (hint: see page 3 of lecture 7).
︡66298d77-dd15-43d9-b749-b047f6bccfc1︡{"done":true,"md":"### Q8. Find elements of each possible order in the class group (hint: see page 3 of lecture 7)."}
︠67d242fe-4309-4c4c-8e28-22f9016d2ee3i︠
%md
### Q9. Decompose the quadratic form (7, 3, 49) over the generators of the class group (hint: can be solved by inspection).
︡7a18f38b-9977-4797-9f7f-3c1994705902︡{"done":true,"md":"### Q9. Decompose the quadratic form (7, 3, 49) over the generators of the class group (hint: can be solved by inspection)."}
︠448d27d8-040d-4542-b903-de2ee3b8fd4bi︠

%md
### Q10. The class group is cyclic, so from Q8 we should have one generator of order equal to the class number, call it g. For the prime forms f above 7, 11, and 19, solve the discrete log problem f = g^x.

︡46d0f85b-6143-490c-83aa-6e5cccbfb6a7︡{"done":true,"md":"### Q10. The class group is cyclic, so from Q8 we should have one generator of order equal to the class number, call it g. For the prime forms f above 7, 11, and 19, solve the discrete log problem f = g^x."}
︠d6202acc-d2bb-4ad0-accb-dad6dcf27962︠









